import asyncio
import aiohttp
import hmac
import hashlib
import ssl
import logging
import datetime
import json
from typing import List, Optional

from binance.Pair import Pair
from binance.subscriptions import Subscription, SubscriptionMgr
from binance import enums
from binance.Timer import Timer
from binance.BinanceException import BinanceException

LOG = logging.getLogger(__name__)

class BinanceClient(object):
	REST_API_URI = "https://api.binance.com/api/v3/"

	def __init__(self, certificate_path : str = None, api_key : str = None, sec_key : str = None,
	             api_trace_log : bool = False) -> None:
		self.api_key = api_key
		self.sec_key = sec_key
		self.api_trace_log = api_trace_log

		self.rest_session = None

		self.ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
		self.ssl_context.load_verify_locations(certificate_path)

		self.subscription_sets = []

	async def ping(self) -> dict:
		return await self._create_get("ping")

	async def get_exchange_info(self) -> dict:
		return await self._create_get("exchangeInfo")

	async def get_time(self) -> dict:
		return await self._create_get("time")

	async def get_orderbook(self, pair : Pair, limit : enums.DepthLimit = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": pair,
		})

		if limit:
			params['limit'] = limit.value

		return await self._create_get("depth", params = params)

	async def get_trades(self, pair : Pair, limit : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": pair,
			"limit": limit
		})

		return await self._create_get("trades", params = params)

	async def get_historical_trades(self, pair : Pair, limit : int = None, from_id : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": pair,
			"limit": limit,
			"fromId": from_id
		})

		return await self._create_get("historicalTrades", params = params, headers = self._get_header_api_key())

	async def get_aggregate_trades(self, pair : Pair, limit : int = None, from_id : int = None, start_tmstmp_ms : int = None, end_tmstmp_ms : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": pair,
			"limit": limit,
			"fromId": from_id,
			"startTime": start_tmstmp_ms,
			"endTime": end_tmstmp_ms
		})

		return await self._create_get("aggTrades", params = params)

	async def get_candelsticks(self, pair : Pair, limit : int = None, interval : enums.CandelstickInterval = None, start_tmstmp_ms : int = None, end_tmstmp_ms : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": pair,
			"limit": limit,
			"startTime": start_tmstmp_ms,
			"endTime": end_tmstmp_ms
		})

		if interval:
			params['interval'] = interval.value

		return await self._create_get("klines", params = params)

	async def get_average_price(self, pair : Pair) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": pair
		})

		return await self._create_get("avgPrice", params = params)

	async def get_24h_price_ticker(self, pair : Pair = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": pair,
		})

		return await self._create_get("ticker/24hr", params = params)

	async def get_price_ticker(self, pair : Pair = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": pair,
		})

		return await self._create_get("ticker/price", params = params)

	async def get_best_orderbook_ticker(self, pair : Optional[Pair] = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": pair,
		})

		return await self._create_get("ticker/bookTicker", headers = self._get_header_api_key(), params = params)

	async def create_order(self, pair : Pair, side : enums.OrderSide, type : enums.OrderType,
	                             quantity : str,
	                             price : str = None,
								 stop_price: str = None,
	                             quote_order_quantity: str = None,
	                             time_in_force: enums.TimeInForce = None,
	                             new_client_order_id : str = None,
	                             iceberg_quantity : str = None,
	                             new_order_response_type : enums.OrderResponseType = None,
	                             recv_window_ms : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": str(pair),
			"side": side.value,
			"type": type.value,
			"quantity": quantity,
			"quoteOrderQty": quote_order_quantity,
			"price": price,
			"stopPrice": stop_price,
			"newClientOrderId": new_client_order_id,
			"icebergQty": iceberg_quantity,
			"recvWindow": recv_window_ms,
			"timestamp": self._get_current_timestamp_ms()
		})

		if time_in_force:
			params['timeInForce'] = time_in_force.value

		if new_order_response_type:
			params['newOrderRespType'] = new_order_response_type.value

		return await self._create_post("order", params = params, headers = self._get_header_api_key(), signed = True)

	async def create_test_order(self, pair : Pair, side : enums.OrderSide, type : enums.OrderType,
	                             quantity : str,
	                             price : str = None,
								 stop_price: str = None,
	                             quote_order_quantity: str = None,
	                             time_in_force: enums.TimeInForce = None,
	                             new_client_order_id : str = None,
	                             iceberg_quantity : str = None,
	                             new_order_response_type : enums.OrderResponseType = None,
	                             recv_window_ms : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": str(pair),
			"side": side.value,
			"type": type.value,
			"quantity": quantity,
			"quoteOrderQty": quote_order_quantity,
			"price": price,
			"stopPrice": stop_price,
			"newClientOrderId": new_client_order_id,
			"icebergQty": iceberg_quantity,
			"recvWindow": recv_window_ms,
			"timestamp": self._get_current_timestamp_ms()
		})

		if time_in_force:
			params['timeInForce'] = time_in_force.value

		if new_order_response_type:
			params['newOrderRespType'] = new_order_response_type.value

		return await self._create_post("order/test", params = params, headers = self._get_header_api_key(), signed = True)

	async def get_order(self, pair : Pair, order_id : int = None, orig_client_order_id : int = None, recv_window_ms : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": str(pair),
			"orderId": order_id,
			"origClientOrderId": orig_client_order_id,
			"recvWindow": recv_window_ms,
			"timestamp": self._get_current_timestamp_ms()
		})

		return await self._create_get("order", params = params, headers = self._get_header_api_key(), signed = True)

	async def cancel_order(self, pair : Pair, order_id : str = None, orig_client_order_id : str = None,
	                       new_client_order_id : str = None, recv_window_ms : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": str(pair),
			"orderId": order_id,
			"origClientOrderId": orig_client_order_id,
			"newClientOrderId": new_client_order_id,
			"recvWindow": recv_window_ms,
			"timestamp": self._get_current_timestamp_ms()
		})

		return await self._create_delete("order", params = params, headers = self._get_header_api_key(), signed = True)

	async def get_open_orders(self, pair : Pair = None, recv_window_ms : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": str(pair),
			"recvWindow": recv_window_ms,
			"timestamp": self._get_current_timestamp_ms()
		})

		return await self._create_get("openOrders", params = params, headers = self._get_header_api_key(), signed = True)

	async def get_all_orders(self, pair : Pair, order_id : int = None, limit : int = None, start_tmstmp_ms : int = None, end_tmstmp_ms : int = None, recv_window_ms : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": str(pair),
			"orderId": order_id,
			"startTime": start_tmstmp_ms,
			"endTime": end_tmstmp_ms,
			"limit": limit,
			"recvWindow": recv_window_ms,
			"timestamp": self._get_current_timestamp_ms()
		})

		return await self._create_get("allOrders", params = params, headers = self._get_header_api_key(), signed = True)

	async def create_oco_order(self, pair : Pair, side : enums.OrderSide,
	                           quantity : str,
	                           price: str,
	                           stop_price: str,
	                           limit_client_order_id: str = None,
	                           list_client_order_id: str = None,
	                           limit_iceberg_quantity: str = None,
	                           stop_client_order_id: str = None,
	                           stop_limit_price: str = None,
	                           stop_iceberg_quantity: str = None,
	                           stop_limit_time_in_force: enums.TimeInForce = None,
	                           new_order_response_type : enums.OrderResponseType = None,
	                           recv_window_ms : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": str(pair),
			"side": side.value,
			"quantity": quantity,
			"listClientOrderId": list_client_order_id,
			"limitClientOrderId": limit_client_order_id,
			"price": price,
			"stopClientOrderId": stop_client_order_id,
			"stopPrice": stop_price,
			"stopLimitPrice": stop_limit_price,
			"stopIcebergQty": stop_iceberg_quantity,
			"limitIcebergQty": limit_iceberg_quantity,
			"recvWindow": recv_window_ms,
			"timestamp": self._get_current_timestamp_ms()
		})

		if stop_limit_time_in_force:
			params['stopLimitTimeInForce'] = stop_limit_time_in_force.value

		if new_order_response_type:
			params['newOrderRespType'] = new_order_response_type.value

		return await self._create_post("order/oco", params = params, headers = self._get_header_api_key(), signed = True)

	async def cancel_oco_order(self, pair : Pair, order_list_id : str = None, list_client_order_id : str = None,
	                       new_client_order_id : str = None, recv_window_ms : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": str(pair),
			"orderListId": order_list_id,
			"listClientOrderId": list_client_order_id,
			"newClientOrderId": new_client_order_id,
			"recvWindow": recv_window_ms,
			"timestamp": self._get_current_timestamp_ms()
		})

		return await self._create_delete("orderList", params = params, headers = self._get_header_api_key(), signed = True)

	async def get_oco_order(self, order_list_id : int = None, orig_client_order_id : int = None, recv_window_ms : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"orderListId": order_list_id,
			"origClientOrderId": orig_client_order_id,
			"recvWindow": recv_window_ms,
			"timestamp": self._get_current_timestamp_ms()
		})

		return await self._create_get("orderList", params = params, headers = self._get_header_api_key(), signed = True)

	async def get_all_oco_orders(self, from_id : int = None, limit : int = None, start_tmstmp_ms : int = None, end_tmstmp_ms : int = None, recv_window_ms : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"fromId": from_id,
			"startTime": start_tmstmp_ms,
			"endTime": end_tmstmp_ms,
			"limit": limit,
			"recvWindow": recv_window_ms,
			"timestamp": self._get_current_timestamp_ms()
		})

		return await self._create_get("allOrderList", params = params, headers = self._get_header_api_key(), signed = True)

	async def get_open_oco_orders(self, recv_window_ms : int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"recvWindow": recv_window_ms,
			"timestamp": self._get_current_timestamp_ms()
		})

		return await self._create_get("openOrderList", params = params, headers = self._get_header_api_key(), signed = True)

	async def get_account(self, recv_window_ms: Optional[int] = None) -> dict:
		params = BinanceClient._clean_request_params({
			"recvWindow": recv_window_ms,
			"timestamp": self._get_current_timestamp_ms()
		})

		return await self._create_get("account", headers = self._get_header_api_key(), params = params, signed = True)

	async def get_account_trades(self, pair: Pair, limit: int = None, from_id: int = None,
	                               start_tmstmp_ms: int = None, end_tmstmp_ms: int = None) -> dict:
		params = BinanceClient._clean_request_params({
			"symbol": pair,
			"limit": limit,
			"fromId": from_id,
			"startTime": start_tmstmp_ms,
			"endTime": end_tmstmp_ms,
			"timestamp": self._get_current_timestamp_ms()
		})

		return await self._create_get("myTrades", params = params, headers = self._get_header_api_key(), signed = True)

	async def get_listen_key(self):
		return await self._create_post("userDataStream", headers = self._get_header_api_key())

	def compose_subscriptions(self, subscriptions : List[Subscription]) -> None:
		self.subscription_sets.append(subscriptions)

	async def start_subscriptions(self) -> None:
		if len(self.subscription_sets):
			done, pending = await asyncio.wait(
				[asyncio.create_task(SubscriptionMgr(subscriptions, self.api_key, self.ssl_context).run()) for subscriptions in self.subscription_sets],
				return_when = asyncio.FIRST_EXCEPTION
			)
			for task in done:
				try:
					task.result()
				except Exception as e:
					LOG.exception(f"Unrecoverable exception occurred while processing messages: {e}")
					LOG.info("All websockets scheduled for shutdown")
					for task in pending:
						if not task.cancelled():
							task.cancel()
		else:
			raise Exception("ERROR: There are no subscriptions to be started.")

	async def close(self) -> None:
		session = self._get_rest_session()
		if session is not None:
			await session.close()

	async def _create_get(self, resource : str, params : dict = None, headers : dict = None, signed : bool = False) -> dict:
		return await self._create_rest_call(enums.RestCallType.GET, resource, None, params, headers, signed)

	async def _create_post(self, resource : str, data : dict = None, params : dict = None, headers : dict = None, signed : bool = False) -> dict:
		return await self._create_rest_call(enums.RestCallType.POST, resource, data, params, headers, signed)

	async def _create_delete(self, resource : str, params : dict = None, headers : dict = None, signed : bool = False) -> dict:
		return await self._create_rest_call(enums.RestCallType.DELETE, resource, None, params, headers, signed)

	async def _create_put(self, resource : str, params : dict = None, headers : dict = None, signed : bool = False) -> dict:
		return await self._create_rest_call(enums.RestCallType.PUT, resource, None, params, headers, signed)

	async def _create_rest_call(self, rest_call_type : enums.RestCallType, resource : str, data : dict = None, params : dict = None, headers : dict = None, signed : bool = False) -> dict:
		with Timer('RestCall'):
			# add signature into parameters
			if signed:
				params = {} if params is None else params
				params['signature'] = self._get_signature(params, data)

			if rest_call_type == enums.RestCallType.GET:
				rest_call = self._get_rest_session().get(BinanceClient.REST_API_URI + resource, json = data, params = params, headers = headers, ssl = self.ssl_context)
			elif rest_call_type == enums.RestCallType.POST:
				rest_call = self._get_rest_session().post(BinanceClient.REST_API_URI + resource, json = data, params = params, headers = headers, ssl = self.ssl_context)
			elif rest_call_type == enums.RestCallType.DELETE:
				rest_call = self._get_rest_session().delete(BinanceClient.REST_API_URI + resource, json = data, params = params, headers = headers, ssl = self.ssl_context)
			elif rest_call_type == enums.RestCallType.PUT:
				rest_call = self._get_rest_session().put(BinanceClient.REST_API_URI + resource, json = data, params = params, headers = headers, ssl = self.ssl_context)
			else:
				raise Exception(f"Unsupported REST call type {rest_call_type}.")

			LOG.debug(f"> rest type [{rest_call_type.name}], resource [{resource}], params [{params}], headers [{headers}], data [{data}]")
			async with rest_call as response:
				status_code = response.status
				response_body = await response.text()

				LOG.debug(f"<: status [{status_code}], response [{response_body}]")

				if str(status_code)[0] != '2':
					raise BinanceException(f"<: status [{status_code}], response [{response_body}]")

				if len(response_body) > 0:
					response_body = json.loads(response_body)

				return {
					"status_code": status_code,
					"response": response_body
				}

	def _get_rest_session(self) -> aiohttp.ClientSession:
		if self.rest_session is not None:
			return self.rest_session

		if self.api_trace_log:
			trace_config = aiohttp.TraceConfig()
			trace_config.on_request_start.append(BinanceClient._on_request_start)
			trace_config.on_request_end.append(BinanceClient._on_request_end)
			trace_configs = [trace_config]
		else:
			trace_configs = None

		self.rest_session = aiohttp.ClientSession(trace_configs=trace_configs)

		return self.rest_session

	def _get_header_api_key(self):
		header = {
			'Accept': 'application/json',
			"X-MBX-APIKEY": self.api_key
		}

		return header

	@staticmethod
	def _clean_request_params(params : dict) -> dict:
		res = {}
		for key, value in params.items():
			if value is not None:
				res[key] = str(value)

		return res

	async def _on_request_start(session, trace_config_ctx, params) -> None:
		LOG.debug(f"> Context: {trace_config_ctx}")
		LOG.debug(f"> Params: {params}")

	async def _on_request_end(session, trace_config_ctx, params) -> None:
		LOG.debug(f"< Context: {trace_config_ctx}")
		LOG.debug(f"< Params: {params}")

	@staticmethod
	def _get_current_timestamp_ms() -> int:
		return int(datetime.datetime.now(tz = datetime.timezone.utc).timestamp() * 1000)

	def _get_signature(self, params : dict, data : dict) -> str:
		params_string = ""
		data_string = ""

		if params is not None:
			params_string = '&'.join([f"{key}={val}" for key, val in params.items()])

		if data is not None:
			data_string = '&'.join(["{}={}".format(param[0], param[1]) for param in data])

		m = hmac.new(self.sec_key.encode('utf-8'), (params_string+data_string).encode('utf-8'), hashlib.sha256)
		return m.hexdigest()


import websockets
import json
import logging
import asyncio
from abc import ABC, abstractmethod
from typing import List, Callable, Any

from binance.Pair import Pair

LOG = logging.getLogger(__name__)

class Subscription(ABC):
	def __init__(self, callbacks = None):
		self.callbacks = callbacks

	@abstractmethod
	def get_channel_name(self) -> str:
		pass

	async def initialize(self) -> None:
		pass

	async def process_message(self, response : dict) -> None:
		await self.process_callbacks(response)

	async def process_callbacks(self, response : dict) -> None:
		if self.callbacks is not None:
			await asyncio.gather(*[asyncio.create_task(cb(response)) for cb in self.callbacks])


class SubscriptionMgr(object):
	WEB_SOCKET_URI = "wss://stream.binance.com:9443/"

	SUBSCRIPTION_ID = 0

	def __init__(self, subscriptions : List[Subscription], api_key : str, ssl_context = None):
		self.api_key = api_key
		self.ssl_context = ssl_context

		self.subscriptions = subscriptions

	async def run(self) -> None:
		for subscription in self.subscriptions:
			await subscription.initialize()

		try:
			# main loop ensuring proper reconnection after a graceful connection termination by the remote server
			while True:
				LOG.debug(f"Initiating websocket connection.")
				uri = SubscriptionMgr.WEB_SOCKET_URI + self._create_stream_uri()
				LOG.debug(f"Websocket uri: {uri}")
				async with websockets.connect(uri, ssl = self.ssl_context, ping_interval = None) as websocket:
					subscription_message = self._create_subscription_message()
					LOG.debug(f"> {subscription_message}")
					await websocket.send(json.dumps(subscription_message))

					# start processing incoming messages
					while True:
						response = json.loads(await websocket.recv())
						LOG.debug(f"< {response}")

						if self._is_subscription_confirmation(response):
							LOG.info(f"Subscription confirmed for id: {response['id']}")
						# regular message
						else:
							await self.process_message(response)
		except asyncio.CancelledError:
			LOG.warning(f"Websocket requested to be shutdown.")
		except Exception:
			LOG.error(f"Exception occurred. Websocket will be closed.")
			raise

	def _create_subscription_message(self) -> dict:
		SubscriptionMgr.SUBSCRIPTION_ID += 1

		return {
			"method": "SUBSCRIBE",
			"params": [
				subscription.get_channel_name() for subscription in self.subscriptions
			],
			"id": SubscriptionMgr.SUBSCRIPTION_ID
		}

	def _create_stream_uri(self) -> str:
		return "stream?streams=" + "/".join([subscription.get_channel_name() for subscription in self.subscriptions])

	@staticmethod
	def _is_subscription_confirmation(response):
		if 'result' in response and response['result'] is None:
			return True
		else:
			return False

	async def process_message(self, response : dict) -> None:
		for subscription in self.subscriptions:
			if subscription.get_channel_name() == response["stream"]:
				await subscription.process_message(response["data"])
				break

class BestOrderBookTickerSubscription(Subscription):
	def __init__(self, callbacks : List[Callable[[dict], Any]] = None):
		super().__init__(callbacks)

	def get_channel_name(self):
		return "!bookTicker"

class TradeSubscription(Subscription):
	def __init__(self, pair : Pair, callbacks: List[Callable[[dict], Any]] = None):
		super().__init__(callbacks)

		self.pair = pair

	def get_channel_name(self):
		return str(self.pair).lower() + "@trade"

class AccountSubscription(Subscription):
	def __init__(self, binance_client, callbacks: List[Callable[[dict], Any]] = None):
		super().__init__(callbacks)

		self.binance_client = binance_client
		self.listen_key = None

	async def initialize(self):
		listen_key_response = await self.binance_client.get_listen_key()
		self.listen_key = listen_key_response["response"]["listenKey"]
		LOG.debug(f'Listen key: {self.listen_key}')

	def get_channel_name(self):
		return self.listen_key


import enum

class RestCallType(enum.Enum):
	GET = enum.auto()
	POST = enum.auto()
	DELETE = enum.auto()
	PUT = enum.auto()

class OrderSide(enum.Enum):
	BUY = "BUY"
	SELL = "SELL"

class OrderType(enum.Enum):
	LIMIT = 'LIMIT'
	MARKET = 'MARKET'
	STOP_LOSS = 'STOP_LOSS'
	STOP_LOSS_LIMIT = 'STOP_LOSS_LIMIT'
	TAKE_PROFIT = 'TAKE_PROFIT'
	TAKE_PROFIT_LIMIT = 'TAKE_PROFIT_LIMIT'
	LIMIT_MAKER = 'LIMIT_MAKER'

class DepthLimit(enum.Enum):
	L_5 = '5'
	L_10 = '10'
	L_20 = '20'
	L_50 = '50'
	L_100 = '100'
	L_500 = '500'
	L_1000 = '1000'
	L_5000 = '5000'

class CandelstickInterval(enum.Enum):
	I_1MIN = '1m'
	I_3MIN = '3m'
	I_5MIN = '5m'
	I_15MIN = '15m'
	I_30MIN = '30m'
	I_1H = '1h'
	I_2H = '2h'
	I_4H = '4h'
	I_6H = '6h'
	I_8H = '8h'
	I_12H = '12h'
	I_1D = '1d'
	I_3D = '3d'
	I_1W = '1w'
	I_1MONTH = '1M'

class OrderResponseType(enum.Enum):
	ACT = "ACK"
	RESULT = "RESULT"
	FULL = "FULL"

class TimeInForce(enum.Enum):
	GOOD_TILL_CANCELLED = "GTC"
	IMMEDIATE_OR_CANCELLED = "IOC"
	FILL_OR_KILL = "FOK"

